"""IMF Direction Of Trade Model."""

# pylint: disable=unused-argument

from datetime import datetime
from typing import Any, Optional, Union

from openbb_core.app.model.abstract.error import OpenBBError
from openbb_core.provider.abstract.fetcher import Fetcher
from openbb_core.provider.standard_models.direction_of_trade import (
    DirectionOfTradeData,
    DirectionOfTradeQueryParams,
)
from openbb_core.provider.utils.errors import EmptyDataError

from openbb_imf.utils.dot_helpers import (
    load_country_map,
    load_country_to_code_map,
    validate_countries,
)

dot_indicators_dict = {
    "exports": "TXG_FOB_USD",
    "imports": "TMG_CIF_USD",
    "balance": "TBG_USD",
    "all": "TXG_FOB_USD+TMG_CIF_USD+TBG_USD",
}

dot_titles_map = {
    "TXG_FOB_USD": "Goods, Value of Exports, Free on board (FOB), US Dollars",
    "TMG_CIF_USD": "Goods, Value of Imports, Cost, Insurance, Freight (CIF), US Dollars",
    "TBG_USD": "Goods, Value of Trade Balance, US Dollars",
}


class ImfDirectionOfTradeQueryParams(DirectionOfTradeQueryParams):
    """IMF Direction Of Trade Query Parameters."""

    __json_schema_extra__ = {
        "country": {
            "multiple_items_allowed": True,
            "choices": ["all"] + sorted(list(load_country_to_code_map())),
        },
        "counterpart": {
            "multiple_items_allowed": True,
            "choices": ["all"] + sorted(list(load_country_to_code_map())),
        },
    }


class ImfDirectionOfTradeData(DirectionOfTradeData):
    """IMF Direction Of Trade Data."""


class ImfDirectionOfTradeFetcher(
    Fetcher[ImfDirectionOfTradeQueryParams, list[ImfDirectionOfTradeData]]
):
    """IMF Direction Of Trade Fetcher."""

    @staticmethod
    def transform_query(params: dict[str, Any]) -> ImfDirectionOfTradeQueryParams:
        """Transform query parameters."""
        countries = params.get("country", "")
        countries = countries.split(",") if countries else "all"
        if countries != "all":
            countries = validate_countries(countries)
        counterparts = params.get("counterpart", "")
        counterparts = counterparts.split(",") if params.get("counterpart") else "all"
        if counterparts != "all":
            counterparts = validate_countries(counterparts)
        now = datetime.now().date()

        if countries == "all" and counterparts == "all":
            raise OpenBBError(
                "Both 'country' and 'counterpart' cannot be None, or 'all'."
                + " Please supply lowercase country names or two-letter ISO codes."
            )
        if countries == counterparts:
            raise OpenBBError("The 'country' and 'counterpart' cannot be the same.")

        params["country"] = countries
        params["counterpart"] = counterparts

        if not params.get("end_date"):
            params["end_date"] = now.replace(month=12, day=31).strftime("%Y-%m-%d")

        if (countries == "all" or counterparts == "all") and not params.get(
            "start_date"
        ):
            params["start_date"] = now.replace(year=now.year - 1).strftime("%Y-%m-%d")

        return ImfDirectionOfTradeQueryParams(**params)

    @staticmethod
    async def aextract_data(
        query: ImfDirectionOfTradeQueryParams,
        credentials: Optional[dict[str, str]],
        **kwargs: Any,
    ) -> list[dict]:
        """Extract the data from the IMF API."""
        # pylint: disable=import-outside-toplevel
        from aiohttp.client_exceptions import ContentTypeError  # noqa
        from json import JSONDecodeError
        from openbb_core.provider.utils.helpers import amake_request
        from pandas import to_datetime
        from pandas.tseries import offsets

        start_date = query.start_date
        end_date = query.end_date
        frequency = query.frequency[0].upper()
        country = query.country if query.country != "all" else ""
        counterpart = query.counterpart if query.counterpart != "all" else ""
        indicator = dot_indicators_dict[query.direction]
        # Adjust the dates to the date relative to frequency.
        # The API does not accept arbitrary dates, so we need to adjust them.
        if start_date:
            start_date = to_datetime(start_date)
            if frequency == "Q":
                start_date = offsets.QuarterBegin(startingMonth=1).rollback(start_date)
            elif frequency == "A":
                start_date = offsets.YearBegin().rollback(start_date)
            else:
                start_date = offsets.MonthBegin().rollback(start_date)
            start_date = start_date.strftime("%Y-%m-%d")  # type: ignore

        if end_date:
            end_date = to_datetime(end_date)
            if frequency == "Q":
                end_date = offsets.QuarterEnd().rollforward(end_date)
            elif frequency == "A":
                end_date = offsets.YearEnd().rollforward(end_date)
            else:
                end_date = offsets.MonthEnd().rollforward(end_date)
            end_date = end_date.strftime("%Y-%m-%d")  # type: ignore

        date_range = (  # type: ignore
            f"?startPeriod={start_date}&endPeriod={end_date}"
            if start_date and end_date
            else ""
        )
        base_url = "http://dataservices.imf.org/REST/SDMX_JSON.svc/"
        key = f"CompactData/DOT/{frequency}.{country}.{indicator}.{counterpart}"
        url = f"{base_url}{key}{date_range}"

        try:
            response = await amake_request(url, timeout=20)
        except (JSONDecodeError, ContentTypeError) as e:
            raise OpenBBError(
                "Error fetching data; This might be rate-limiting. Try again later."
            ) from e

        if "ErrorDetails" in response:
            raise OpenBBError(
                f"{response['ErrorDetails'].get('Code')} -> {response['ErrorDetails'].get('Message')}"  # type: ignore
            )

        series = response.get("CompactData", {}).get("DataSet", {}).pop("Series", {})  # type: ignore

        if not series:
            raise OpenBBError(f"No time series data found -> {url} -> {response}")

        # If there is only one series, they ruturn a dict instead of a list.
        if series and isinstance(series, dict):
            series = [series]

        return series

    @staticmethod
    def transform_data(
        query: ImfDirectionOfTradeQueryParams,
        data: list[dict],
        **kwargs: Any,
    ) -> list[ImfDirectionOfTradeData]:
        """Transform the data."""
        # pylint: disable=import-outside-toplevel
        from openbb_imf.utils.constants import UNIT_MULTIPLIERS_MAP  # noqa
        from pandas import Categorical, DataFrame, to_datetime
        from pandas.tseries import offsets

        if not data:
            raise EmptyDataError()

        dot_code_to_country = load_country_map()
        series = data
        results: list = []

        for s in series:
            if "Obs" not in s:
                continue
            meta = {
                k.replace("@", "").lower(): (
                    UNIT_MULTIPLIERS_MAP.get(str(v), v) if k == "@UNIT_MULT" else v
                )
                for k, v in s.items()
                if k != "Obs"
            }
            _symbol = meta.get("indicator", "")
            _title = None

            _data = s.pop("Obs", [])

            if isinstance(_data, dict):
                _data = [_data]

            for d in _data:
                _date = d.pop("@TIME_PERIOD", None)
                val: Union[float, None] = d.pop("@OBS_VALUE", None)
                _ = d.pop("@OBS_STATUS", None)
                val = float(val) if val else None
                if not val:
                    continue

                if _date:
                    offset = (
                        offsets.QuarterEnd
                        if "Q" in _date
                        else (
                            offsets.YearEnd
                            if len(str(_date)) == 4
                            else offsets.MonthEnd
                        )
                    )
                    _date = to_datetime(_date)
                    _date = _date + offset(0)
                    _date = _date.strftime("%Y-%m-%d")
                vals = {
                    k: v
                    for k, v in {
                        "date": _date,
                        "symbol": _symbol,
                        "country": dot_code_to_country.get(
                            meta.get("ref_area"), meta.get("ref_area")
                        ),
                        "counterpart": dot_code_to_country.get(
                            meta.get("counterpart_area"), meta.get("counterpart_area")
                        ),
                        "title": dot_titles_map.get(_symbol),
                        "scale": meta.get("unit_mult"),
                        "value": val,
                    }.items()
                    if v
                }

                if (
                    vals.get("value")
                    and vals.get("date")
                    and vals.get("country") != vals.get("counterpart")
                ):
                    d.update(vals)

            if _data:
                results.extend([d for d in _data if d])

        df = DataFrame(results)
        df["symbol"] = Categorical(
            df["symbol"],
            categories=list(dot_titles_map),
            ordered=True,
        )
        df["country"] = Categorical(
            df["country"],
            categories=sorted(df.country.unique().tolist()),
            ordered=True,
        )
        df["counterpart"] = Categorical(
            df["counterpart"],
            categories=sorted(df.counterpart.unique().tolist()),
            ordered=True,
        )
        df = df.sort_values(by=["date", "country", "counterpart"])

        return [
            ImfDirectionOfTradeData.model_validate(r)
            for r in df.to_dict(orient="records")
        ]



"""IMF Available Indicators."""

# pylint: disable=unused-argument

from typing import Any, Dict, List, Optional

from openbb_core.app.model.abstract.error import OpenBBError
from openbb_core.provider.abstract.fetcher import Fetcher
from openbb_core.provider.standard_models.available_indicators import (
    AvailableIndicatorsData,
    AvailableIndicesQueryParams,
)
from pydantic import Field


class ImfAvailableIndicatorsQueryParams(AvailableIndicesQueryParams):
    """IMF Available Indicators Query Parameters."""


class ImfAvailableIndicatorsData(AvailableIndicatorsData):
    """IMF Available Indicators Data."""

    __alias_dict__ = {
        "symbol_root": "parent",
        "description": "title",
    }

    table: Optional[str] = Field(
        default=None,
        description="The name of the table associated with the symbol.",
    )
    level: Optional[int] = Field(
        default=None,
        description="The indentation level of the data, relative to the table and symbol_root",
    )
    order: Optional[int] = Field(
        default=None,
        description="Order of the data, relative to the table.",
    )


class ImfAvailableIndicatorsFetcher(
    Fetcher[ImfAvailableIndicatorsQueryParams, List[ImfAvailableIndicatorsData]]
):
    """IMF Available Indicators Fetcher."""

    @staticmethod
    def transform_query(params: Dict[str, Any]) -> ImfAvailableIndicatorsQueryParams:
        """Transform the query."""
        return ImfAvailableIndicatorsQueryParams(**params)

    @staticmethod
    def extract_data(
        query: ImfAvailableIndicatorsQueryParams,
        credentials: Optional[Dict[str, Any]] = None,
        **kwargs: Any,
    ) -> List[Dict]:
        """Fetch the data."""
        # pylint: disable=import-outside-toplevel
        from openbb_imf.utils.irfcl_helpers import load_irfcl_symbols

        try:
            irfcl_symbols = load_irfcl_symbols()
        except OpenBBError as e:
            raise OpenBBError(f"Failed to load IMF IRFCL symbols: {e}") from e

        records: List = []
        for key, value in irfcl_symbols.items():
            value["symbol"] = key
            records.append(value)

        return records

    @staticmethod
    def transform_data(
        query: ImfAvailableIndicatorsQueryParams,
        data: List[Dict],
        **kwargs: Any,
    ) -> List[ImfAvailableIndicatorsData]:
        """Transform the data."""
        return [ImfAvailableIndicatorsData.model_validate(d) for d in data]



"""Direction Of Trade Utilities."""

from openbb_core.app.model.abstract.error import OpenBBError


def load_country_map() -> dict:
    """Load IMF IRFCL country map."""
    # pylint: disable=import-outside-toplevel
    import json  # noqa
    from json.decoder import JSONDecodeError
    from pathlib import Path

    try:
        country_map_file = (
            Path(__file__).parents[1].joinpath("assets", "imf_country_map.json")
        )
        with country_map_file.open(encoding="utf-8") as file:
            country_map_dict = json.load(file)
    except (FileNotFoundError, JSONDecodeError) as e:
        raise OpenBBError(f"Failed to load IMF DOT country map: {e}") from e

    return country_map_dict


def load_country_to_code_map() -> dict:
    """Load a map of lowercase country name to 2-letter ISO symbol."""
    # pylint: disable=import-outside-toplevel
    from warnings import warn  # noqa

    try:
        return {
            (
                "euro_area"
                if k == "U2"
                else v.lower()
                .replace(".", "")
                .replace(",", "")
                .replace(":", "")
                .split("(")[0]
                .strip()
                .replace(" ", "_")
            ): k
            for k, v in load_country_map().items()
            if not k.startswith("1C_ALL")
            and "Report" not in v
            and k not in ("GW", "_X")
        }
    except OpenBBError:
        warn(f"Failed to load country to code map. -> {OpenBBError}")
        return {}


def validate_countries(countries: list[str]) -> str:
    """Validate the list of countries."""
    # pylint: disable=import-outside-toplevel
    from warnings import warn

    try:
        country_to_code_map = load_country_to_code_map()
    except OpenBBError as e:
        raise OpenBBError(f"Failed to load country to code map: {e}") from e
    if not country_to_code_map:
        raise OpenBBError("Failed to load country to code map.")

    new_countries: list = []

    for country in countries:
        if country == "all":
            return country
        if country in country_to_code_map:
            new_countries.append(country_to_code_map.get(country))
        elif country.upper() in country_to_code_map.values():
            new_countries.append(country.upper())
        elif country.lower() == "eu":
            new_countries.append(country_to_code_map.get("european_union"))
        elif country.lower() == "ea":
            new_countries.append(country_to_code_map.get("euro_area"))
        else:
            warn(f"Invalid country {country}")
            continue

    if not new_countries:
        raise OpenBBError(
            f"No valid countries found. Please choose from {list(country_to_code_map)}"
        )

    return "+".join(new_countries)



"""OpenBB IMF Provider Module."""

from openbb_core.provider.abstract.provider import Provider

from openbb_imf.models.available_indicators import ImfAvailableIndicatorsFetcher
from openbb_imf.models.direction_of_trade import ImfDirectionOfTradeFetcher
from openbb_imf.models.economic_indicators import ImfEconomicIndicatorsFetcher

imf_provider = Provider(
    name="imf",
    website="https://datahelp.imf.org/knowledgebase/articles/667681-using-json-restful-web-service",
    description="This provider allows you to access International Monetary Fund data through the IMF Public Data API.",
    fetcher_dict={
        "AvailableIndicators": ImfAvailableIndicatorsFetcher,
        "DirectionOfTrade": ImfDirectionOfTradeFetcher,
        "EconomicIndicators": ImfEconomicIndicatorsFetcher,
    },
    repr_name="International Monetary Fund (IMF) Public Data API",
)



"""IMF IRFCL Data Set Helpers."""

from typing import Dict, List, Optional


def load_irfcl_symbols() -> Dict:
    """Load IMF IRFCL symbols."""
    # pylint: disable=import-outside-toplevel
    import json  # noqa
    from json.decoder import JSONDecodeError
    from pathlib import Path
    from openbb_core.app.model.abstract.error import OpenBBError

    try:
        symbols_file = Path(__file__).parents[1].joinpath("assets", "imf_symbols.json")
        with symbols_file.open(encoding="utf-8") as file:
            symbols = json.load(file)
    except (FileNotFoundError, JSONDecodeError) as e:
        raise OpenBBError(f"Failed to load IMF IRFCL symbols: {e}") from e

    return {k: v for k, v in symbols.items() if v["dataset"] == "IRFCL"}


def load_country_map() -> Dict:
    """Load IMF IRFCL country map."""
    # pylint: disable=import-outside-toplevel
    import json  # noqa
    from json.decoder import JSONDecodeError
    from pathlib import Path
    from openbb_core.app.model.abstract.error import OpenBBError

    try:
        country_map_file = (
            Path(__file__).parents[1].joinpath("assets", "imf_country_map.json")
        )
        with country_map_file.open(encoding="utf-8") as file:
            country_map_dict = json.load(file)
    except (FileNotFoundError, JSONDecodeError) as e:
        raise OpenBBError(f"Failed to load IMF IRFCL country map: {e}") from e

    return {
        k: v.split(",")[0].split("_(")[0]
        for k, v in country_map_dict.items()
        if len(k) == 2
        and k[0] not in ("5", "1", "7")
        and k not in ("X0", "R1", "GW", "F1", "F6")
    }


def load_country_to_code_map() -> Dict:
    """Load a map of lowercase country name to 2-letter ISO symbol."""
    return {
        (
            "euro_area"
            if k == "U2"
            else v.lower()
            .replace(" ", "_")
            .replace("`", "")
            .split(",")[0]
            .split("_(")[0]
        ): k
        for k, v in load_country_map().items()
    }


def validate_countries(countries) -> str:
    """Validate the country and convert to a 2-letter ISO country code."""
    # pylint: disable=import-outside-toplevel
    from warnings import warn  # noqa
    from openbb_core.app.model.abstract.error import OpenBBError

    country_map_dict = load_country_to_code_map()

    if isinstance(countries, str):
        countries = countries.split(",")
    elif not isinstance(countries, list):
        raise OpenBBError("Invalid countries list.")

    new_countries: List = []

    if "all" in countries or "ALL" in countries:
        return "all"

    for country in countries:
        if country.lower() not in country_map_dict and country.upper() not in list(
            country_map_dict.values()
        ):
            warn(f"Invalid IMF IRFCL country: {country}")
            continue

        if country.upper() in list(country_map_dict.values()):
            new_countries.append(country.upper())
        else:
            new_countries.append(country_map_dict.get(country, country).upper())

    new_countries = [c for c in new_countries if c]

    if not new_countries:
        raise OpenBBError("No valid countries found in the supplied list.")

    return ",".join(new_countries)


def validate_symbols(symbols) -> str:
    """Validate the IMF IRFCL symbols."""
    # pylint: disable=import-outside-toplevel
    from warnings import warn  # noqa
    from openbb_core.app.model.abstract.error import OpenBBError
    from openbb_imf.utils.constants import IRFCL_PRESET

    irfcl_symbols = load_irfcl_symbols()

    if isinstance(symbols, str):
        symbols = symbols.split(",")
    elif not isinstance(symbols, list):
        raise OpenBBError("Invalid symbols list.")

    if "IRFCL" in symbols or "all" in symbols:
        return "all"

    new_symbols: List = []

    for symbol in symbols:
        if symbol in IRFCL_PRESET:
            return IRFCL_PRESET[symbol].replace(",", "+")
        if symbol.upper() not in irfcl_symbols:
            warn(f"Invalid IMF IRFCL symbol: {symbol}")
        new_symbols.append(symbol.upper())

    return "+".join(new_symbols) if len(new_symbols) > 1 else new_symbols[0]


# We use this as a helper to allow future expansion of the supported IMF indicators.
# Each database has its own nuances with URL construction and schemas.
# In the future, we will check the supported symbols map and direct the call
# to the appropriate dtatabase's function.


# pylint: disable=too-many-branches,too-many-statements,too-many-locals
async def _get_irfcl_data(**kwargs) -> List[Dict]:
    """Get IMF IRFCL data.
    This function is not intended to be called directly,
    but through the `ImfEconomicIndicatorsFetcher` class.
    """
    # pylint: disable=import-outside-toplevel
    from aiohttp.client_exceptions import ContentTypeError  # noqa
    from json.decoder import JSONDecodeError
    from openbb_core.provider.utils.helpers import amake_request
    from openbb_core.app.model.abstract.error import OpenBBError
    from openbb_imf.utils import constants
    from pandas import to_datetime
    from pandas.tseries import offsets

    countries = kwargs.get("country", "")
    countries = (
        "" if countries == "all" else countries.replace(",", "+") if countries else ""
    )

    frequency = constants.FREQUENCY_DICT.get(kwargs.get("frequency", "quarter"), "Q")
    sector = kwargs.get("sector", "monetary_authorities")
    sector = constants.REF_SECTORS_DICT.get(sector, "")
    start_date = kwargs.get("start_date", "")
    end_date = kwargs.get("end_date", "")

    # Adjust the dates to the date relative to frequency.
    # The API does not accept arbitrary dates, so we need to adjust them.
    if start_date:
        start_date = to_datetime(start_date)
        if frequency == "Q":
            # offset = offsets.QuarterBegin(startingMonth=1)
            # start_date = start_date + offset
            start_date = offsets.QuarterBegin(startingMonth=1).rollback(start_date)
        elif frequency == "A":
            start_date = offsets.YearBegin().rollback(start_date)
        else:
            start_date = offsets.MonthBegin().rollback(start_date)
        start_date = start_date.strftime("%Y-%m-%d")

    if end_date:
        end_date = to_datetime(end_date)
        if frequency == "Q":
            end_date = offsets.QuarterEnd().rollforward(end_date)
        elif frequency == "A":
            end_date = offsets.YearEnd().rollforward(end_date)
        else:
            end_date = offsets.MonthEnd().rollforward(end_date)
        end_date = end_date.strftime("%Y-%m-%d")

    indicator = kwargs.get("symbol")
    indicator = validate_symbols(indicator) if indicator else ""
    indicator = "" if indicator == "all" else indicator

    if not indicator and not countries:
        raise OpenBBError("Country is required when returning the complete dataset.")

    date_range = (
        f"?startPeriod={start_date}&endPeriod={end_date}"
        if start_date and end_date
        else ""
    )
    base_url = "http://dataservices.imf.org/REST/SDMX_JSON.svc/"
    key = f"CompactData/IRFCL/{frequency}.{countries}.{indicator}.{sector}"
    url = f"{base_url}{key}{date_range}"

    try:
        response = await amake_request(url, timeout=20)
    except (JSONDecodeError, ContentTypeError) as e:
        raise OpenBBError(
            "Error fetching data; This might be rate-limiting. Try again later."
        ) from e

    if "ErrorDetails" in response:
        raise OpenBBError(
            f"{response['ErrorDetails'].get('Code')} -> {response['ErrorDetails'].get('Message')}"  # type: ignore
        )

    series = response.get("CompactData", {}).get("DataSet", {}).pop("Series", {})  # type: ignore

    if not series:
        raise OpenBBError(f"No time series data found -> {response}")

    # If there is only one series, they ruturn a dict instead of a list.
    if series and isinstance(series, dict):
        series = [series]

    data: List = []
    all_symbols = load_irfcl_symbols()
    country_map_dict = {
        v: k.replace("_", " ").title().replace("Ecb", "ECB")
        for k, v in load_country_to_code_map().items()
    }
    # Iterate over the series to extract observations and map the metadata.
    for s in series:
        if "Obs" not in s:
            continue
        meta = {
            k.replace("@", "").lower(): (
                constants.UNIT_MULTIPLIERS_MAP.get(str(v), v)
                if k == "@UNIT_MULT"
                else v
            )
            for k, v in s.items()
            if k != "Obs"
        }
        _symbol = meta.get("indicator")
        _parent: Optional[str] = None
        _order: Optional[str] = None
        _level: Optional[str] = None
        _table: Optional[str] = None
        _title: Optional[str] = None

        if _symbol in all_symbols:
            _table = all_symbols.get(_symbol, {}).get("table")
            _parent = all_symbols.get(_symbol, {}).get("parent", "")
            _order = all_symbols.get(_symbol, {}).get("order", "")
            _level = all_symbols.get(_symbol, {}).get("level", "")
            _title = all_symbols.get(_symbol, {}).get("title", "")

        _data = s.pop("Obs", [])

        if isinstance(_data, dict):
            _data = [_data]

        for d in _data:
            _date = d.pop("@TIME_PERIOD", None)
            val = d.pop("@OBS_VALUE", None)

            if not val:
                continue

            if _date:
                offset = (
                    offsets.QuarterEnd
                    if "Q" in _date
                    else offsets.YearEnd if len(str(_date)) == 4 else offsets.MonthEnd
                )
                _date = to_datetime(_date)
                _date = _date + offset(0)
                _date = _date.strftime("%Y-%m-%d")
            vals = {
                k: v
                for k, v in {
                    "date": _date,
                    "table": _table,
                    "symbol": _symbol,
                    "parent": _parent,
                    "order": _order,
                    "level": _level,
                    "country": country_map_dict.get(
                        meta.get("ref_area"), meta.get("ref_area")
                    ),
                    "reference_sector": constants.SECTOR_MAP.get(
                        meta.get("ref_sector"), meta.get("ref_sector")
                    ),
                    "title": _title,
                    "value": float(val) if val else None,
                    "scale": meta.get("unit_mult"),
                }.items()
                if v
            }

            if vals.get("value") and vals.get("date"):
                d.update(vals)

        if _data:
            data.extend([d for d in _data if d])

    if not data:
        raise OpenBBError("No data found.")

    return data



"""IMF Economic Indicators Model."""

# pylint: disable=unused-argument

from datetime import datetime
from typing import Any, Dict, List, Literal, Optional

from openbb_core.app.model.abstract.error import OpenBBError
from openbb_core.provider.abstract.fetcher import Fetcher
from openbb_core.provider.standard_models.economic_indicators import (
    EconomicIndicatorsData,
    EconomicIndicatorsQueryParams,
)
from openbb_core.provider.utils.descriptions import QUERY_DESCRIPTIONS
from openbb_core.provider.utils.errors import EmptyDataError
from pydantic import Field

from openbb_imf.utils.constants import IRFCL_PRESET, IRFCL_TABLES
from openbb_imf.utils.irfcl_helpers import (
    load_country_to_code_map,
    load_irfcl_symbols,
    validate_countries,
)


class ImfEconomicIndicatorsQueryParams(EconomicIndicatorsQueryParams):
    """IMF Economic Indicators Query."""

    __json_schema_extra__ = {
        "symbol": {
            "multiple_items_allowed": True,
        },
        "country": {
            "multiple_items_allowed": True,
            "choices": ["all"] + list(list(load_country_to_code_map())),
        },
        "frequency": {
            "choices": ["annual", "quarter", "month"],
        },
    }
    symbol: str = Field(
        default="irfcl_top_lines",
        description=QUERY_DESCRIPTIONS.get("symbol", "")
        + " Use `available_indicators()` to get the list of available symbols."
        + " Use 'IRFCL' to get all the data from the set of indicators."
        + " Complete tables are available only by single country, and are keyed as described below."
        + " The default is 'irfcl_top_lines'. Available presets not listed in `available_indicators()` are:"
        + """\n
        'IRFCL': All the data from the set of indicators. Not compatible with multiple countries.
        'irfcl_top_lines': The default, top line items from the IRFCL data. Compatible with multiple countries.
        'reserve_assets_and_other_fx_assets': Table I of the IRFCL data. Not compatible with multiple countries.
        'predetermined_drains_on_fx_assets': Table II of the IRFCL data. Not compatible with multiple countries.
        'contingent_drains_fx_assets': Table III of the IRFCL data. Not compatible with multiple countries.
        'memorandum_items': The memorandum items table of the IRFCL data. Not compatible with multiple countries.
        'gold_reserves': Gold reserves as value in USD and Fine Troy Ounces. Compatible with multiple countries.
        'derivative_assets': Net derivative assets as value in USD. Compatible with multipile countries.
    """,
    )
    frequency: Literal["annual", "quarter", "month"] = Field(
        default="quarter",
        description="Frequency of the data.",
    )


class ImfEconomicIndicatorsData(EconomicIndicatorsData):
    """IMF Economic Indicators Data."""

    __alias_dict__ = {
        "symbol_root": "parent",
    }

    scale: Optional[str] = Field(
        default=None,
        description="The scale of the value.",
    )
    table: Optional[str] = Field(
        default=None,
        description="The name of the table associated with the symbol.",
    )
    level: Optional[int] = Field(
        default=None,
        description="The indentation level of the data, relative to the table and symbol_root",
    )
    order: Optional[int] = Field(
        default=None,
        description="Order of the data, relative to the table.",
    )
    reference_sector: Optional[str] = Field(
        default=None,
        description="The reference sector for the data.",
    )
    title: Optional[str] = Field(
        default=None,
        description="The title of the series associated with the symbol.",
    )


class ImfEconomicIndicatorsFetcher(
    Fetcher[ImfEconomicIndicatorsQueryParams, List[ImfEconomicIndicatorsData]]
):
    """IMF Economic Indicators Fetcher."""

    @staticmethod
    def transform_query(params: Dict[str, Any]) -> ImfEconomicIndicatorsQueryParams:
        """Transform the query."""
        symbols = params.get("symbol", "")
        countries = params.get("country")
        now = datetime.now().date()
        symbols = "IRFCL" if ("all" in symbols or "IRFCL" in symbols) else symbols
        if symbols == "IRFCL" and not (countries or countries == "all"):
            raise OpenBBError(
                f"The selected symbol(s), {params.get('symbol')}, requires a single country for the 'country' parameter."
            )

        if countries:
            params["country"] = validate_countries(countries)

        if symbols and symbols in IRFCL_PRESET:
            params["symbol"] = IRFCL_PRESET[symbols]
            if symbols in IRFCL_TABLES and countries and countries.split(",") > 1:
                raise OpenBBError(
                    f"Symbol '{symbols}' is a table and can only be used with one country."
                )
        elif symbols:
            params["symbol"] = symbols

        if not params.get("start_date") and (not countries or countries == "all"):
            params["start_date"] = now.replace(
                year=now.year - 1, month=1, day=1
            ).strftime("%Y-%m-%d")

        if not params.get("end_date"):
            params["end_date"] = now.replace(month=12, day=31).strftime("%Y-%m-%d")

        if (not symbols or symbols == "all") and not params.get("start_date"):
            params["start_date"] = now.replace(year=now.year - 1).strftime("%Y-%m-%d")

        return ImfEconomicIndicatorsQueryParams(**params)

    @staticmethod
    async def aextract_data(
        query: ImfEconomicIndicatorsQueryParams,
        credentials: Optional[Dict[str, Any]] = None,
        **kwargs: Any,
    ) -> List[Dict]:
        """Extract the data."""
        # pylint: disable = import-outside-toplevel
        from openbb_imf.utils.irfcl_helpers import _get_irfcl_data

        try:
            res = await _get_irfcl_data(**query.model_dump(exclude_none=True))
        except OpenBBError as e:
            raise e from e
        if not res:
            raise EmptyDataError("No results returned for the query.")
        return res

    @staticmethod
    def transform_data(
        query: ImfEconomicIndicatorsQueryParams,
        data: List[Dict],
        **kwargs: Any,
    ) -> List[ImfEconomicIndicatorsData]:
        """Transform the data."""
        # pylint: disable = import-outside-toplevel
        from pandas import Categorical, DataFrame

        if not data:
            raise EmptyDataError("The data is empty.")

        all_symbols = load_irfcl_symbols()
        df = DataFrame(data)

        if df.empty:
            raise EmptyDataError("The data is empty.")

        df = df[df["symbol"].isin(all_symbols)]

        if len(df) == 0:
            raise OpenBBError("The data has a length of 0.")

        df["symbol"] = Categorical(
            df["symbol"],
            categories=all_symbols,
            ordered=True,
        )
        df["parent"] = Categorical(
            df["parent"],
            categories=all_symbols,
            ordered=True,
        )
        df = df.sort_values(
            by=["date", "symbol", "parent", "value"],
            ascending=[True, True, True, False],
        ).reset_index(drop=True)

        return [ImfEconomicIndicatorsData(**d) for d in df.to_dict(orient="records")]
